/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"
#include <unistd.h> 

/************************************************/
void usage(){
  char UsageMessage[] =  "\
Usage: SimAlpha [-option] command\n\
 -d: debug mode\n\
 -a: print simple evaluation result\n\
 -c: print systemcall trace\n\
 -v[num]: verbose mode, prints '.' \
by VERBOSE_INTERVAL\n\
 -f[num]: set fd_adjust to verify simulation\n\
 -t[num][m]: print execution trace\n\
 -e[num][m]: end simulation at num code run. \
m means million\n\
 -i[file_name]: file_name as STDIN input file\n\
";
  printf("%s", UsageMessage); exit(0);
}

/** return 24bit unique instruction ID         **/
/************************************************/
int get_code(INST_TYPE ir){
  int id=0;
  int op6    = (ir>>26) & 0x003f;
  int func7  = (ir>>5 ) & 0x007f;
  int func11 = (ir>>5 ) & 0x07ff;
  int jmp_op = (ir>>14) & 0x0003;

  if(op6<=0x0f || op6>=0x1b){
    id = op6 << 16;
  }
  else{
    switch(op6){
    case 0x10:
    case 0x11:
    case 0x12:
    case 0x13: id = (op6 << 16) | func7;  break;
    case 0x14: id =  op6 << 16;           break;
    case 0x15:
    case 0x16:
    case 0x17:
    case 0x18: id = (op6 << 16) | func11; break;
    case 0x19: id =  op6 << 16;           break;
    case 0x1a: id = (op6 << 16) | jmp_op; break;
    default:
      printf("*** get_code default %x\n", op6);
    }
  }
  return id;
}

/* If IR is not the cmov, return 0.             */
/************************************************/
int cmov_ir(INST_TYPE ir){
  int op = (ir>>26) &  0x3F;
  if(op==OP_INTL || op==OP_FLTL){
    switch(get_code(ir)){
    case _CMOVLBS:
    case _CMOVLBC:
    case _CMOVEQ :
    case _CMOVNE :
    case _CMOVLT : 
    case _CMOVGE : 
    case _CMOVLE : 
    case _CMOVGT : 
    case _FCMOVEQ:
    case _FCMOVNE:
    case _FCMOVLT:
    case _FCMOVGE:
    case _FCMOVLE:
    case _FCMOVGT: return 1;
    }
  }
  return 0;
}

/************************************************/
void house_keeper(instruction *p){
  Env *ev = p->ev;
  if(ev->sc->trace_flag!=0 &&
     ev->sc->trace_flag<=ev->e->retired_inst){
    printf("TC:%lld  ", ev->e->retired_inst);
    printf("pc:%9llx ", p->Cpc);
    printf("ir:%08x  ", p->ir);
    int id = get_code(p->ir);
    printf("ID:%02x.%04x\n", id>>16, id&0xffff);
    printf(" Rav: %016llx ", p->Rav);
    printf("Rbv: %016llx\n", p->Rbv);
    printf(" Imm: %016llx ", p->Imm);
    printf("Adr: %016llx\n", p->Adr);
    printf(" Rcv: %016llx\n", p->Rcv);
  }
  if(ev->sc->debug_flag) ev->deb->debugmode(p);
  if(ev->sc->verbose_interval && 
     (ev->e->retired_inst % 
      ev->sc->verbose_interval)==0){
    write(1, ".", 1);
  }
  if(ev->sc->end_tc!=0 && 
     ev->e->retired_inst>=ev->sc->end_tc) 
    ev->sys->running = 0;

}

/************************************************/
void option_from_file(char **opt, char *name){
  char *buf;
  for(int i=0; i<OPTION_MAX; i++) opt[i]=NULL;
  FILE *fp = fopen(name,"r");
  if(fp!=NULL){
    printf("*** Use %s\n", name);
    for(int i=0; !feof(fp);){
      buf = (char *)malloc(256);
      fscanf(fp, "%s", buf);
      if(buf[0]=='-') opt[i++] = buf;
    }
  }
}

/* C interface                                  */
/************************************************/
int simple_cpu(int job, uint_64t *p){
  static simple_chip *chip;
	
  switch(job){
  case 1:{
    printf("%s %s\n", NAME, VER); 
    static char *opt[OPTION_MAX];
    option_from_file(opt, "SimAlpha_Option.txt");
    chip = new simple_chip(DEF_FILE_NAME, opt);
  }
  break;
  case 2:{
    *(p++) = chip->ev->as->pc;
    int running = chip->step();
    
    for(int i=0; i<31; i++) 
      *(p++)=chip->ev->as->r[i];
    for(int i=0; i<31; i++) 
      *(p++)=chip->ev->as->f[i];
    
    if(running==0) delete chip;
    break;
  }
  default: 
    printf("Warning: %d siple_cpu.\n", job);
  }
  return 0;
}

/* map_s is used in LDS & ITOFS instruction     */
/* input of 8bit and output of 11bit            */
/************************************************/
uint_64t map_s(DATA_TYPE in){
  uint_32t i8 = (uint_32t)((in >> 23) & 0xFF);
  
  /** Next line should be valid, 
      SimpleScalar has a bug ? **/
  // if(i8==0xFF)  return 0x00000000000007FFull;
  if(i8==0x00)  return 0x0000000000000000ull;
  
  if(i8 & 0x80) return 0x400ull | (i8 & 0x7F);
  else          return 0x380ull | (i8 & 0x7F);
}

/************************************************/
architecture_state::architecture_state(){
  for(int i=0; i<32; i++) r[i]=0;
  for(int i=0; i<32; i++) f[i]=0;
}

/*
 * Initialize the registers, data file format.
 * Begin by "\@" to indicate the valid data.
 * /@reg 28 0000000000000000
 * /@reg 29 0000000140023e90                    */
/************************************************/
architecture_state::architecture_state(Env *e){
  ev = e;
  for(int i=0; i<32; i++) r[i]=0;
  for(int i=0; i<32; i++) f[i]=0;

  char *program_name = ev->sc->program_name;
  FILE *fp = fp = fopen(program_name, "r");
  if(fp==NULL){
    printf("Bad file name: %s\n", 
	    program_name);
    exit(1);
  }

 /*** File Format Check ***/
  DATA_TYPE dat;
  char buf[4096];
  fgets(buf, 4096, fp);
  if(strncmp(buf+3, "SimAlpha", 8)){
    printf("%s: Not a SimAlpha data file.\n", 
	   program_name);
    exit(1);
  }
  
  while(!feof(fp)){
    int i;
    fgets(buf, 4096, fp);
    if(*buf=='/' && *(buf+1)=='@'){
      sscanf(buf+6,"%d", &i);
      dat = get_hex(buf+9);
      if(i==32) pc=dat;
      else r[i]=dat;
    }
    if(i==32) break;
  }
  fclose(fp);
}

/************************************************/
uint_64t architecture_state::rr(int n){
  if(n==0) return pc;
  if(n>=1  && n<=32) return r[n-1];
  if(n>=33 && n<=64) return f[n-33];
  printf("Error:architecture_state::rr(%d)\n",n);
  exit(1);
  return 0;
}

/************************************************/
void print_evaluation_result(Env *ev){
  class evaluation_result *e = ev->e;
  struct timeval tp;    
  struct timezone tzp;  

  if(ev->sc->simple_result){
    printf("\n");
    for(int i=0; i<60; i++) printf("=");
    printf("\n=== %s %s\n", NAME, VER);
    printf("=== %6lld million code(%14lld code)\n", 
	   e->retired_inst/1000000,e->retired_inst);
    return;
  }

  gettimeofday(&tp, &tzp);
  uint_64t usec = 
    (uint_64t)(tp.tv_sec - e->tp.tv_sec) 
    * 1000000ull + (uint_64t)tp.tv_usec - 
    (uint_64t)e->tp.tv_usec;
  fflush(stdout);
  fflush(stderr);
  uint_64t minute = (usec / 1000000) / 60;
  uint_64t second = (usec / 1000000) % 60;
  printf("\n");
  for(int i=0; i<60; i++) printf("=");
  printf("\n=== %s %s\n", NAME, VER);
  char buf[256];
  getcwd(buf, 256);
  printf("=== cwd: %s\n", buf);
  printf("=== %6lld million code(%14lld code)", 
	 e->retired_inst/1000000,e->retired_inst);
  printf("  %6.3f MIPS\n", 
	 (double)e->retired_inst/(double)usec);
  printf("=== SimAlpha takes ");
  printf("%5lld min %2lld second",
	 minute, second);
  printf(" (%12lld usec)\n", usec); 
  printf("=== SimAlpha starts at %s", 
	 ctime(&e->time_begin));
  printf("=== Memory %5dKB used",
	 BLOCK_SIZE/1024 * e->used_memory_block);
  printf("(block size %2dKB, %4d blocks)\n",
	 BLOCK_SIZE/1024, e->used_memory_block);
  fflush(stdout);
}

/************************************************/
system_config::system_config(char *program, 
			     char **option){
  memset(this, 0, sizeof(system_config));
  strcpy(program_name, program);
  
  if(option==NULL) return;
  for(int i=0; option[i]!=NULL; i++){
    char opt[1024];
    strcpy(opt, option[i]);
    if(opt[0]!='-') continue;
    switch(opt[1]){
    case 'a': simple_result = 1; break;
    case 'c': syscall_flag  = 1; break;
    case 'z': verify_flag   = 1; break;
    case 'd': 
      debug_flag = atoi(&opt[2]);
      if(debug_flag==0) debug_flag=1;
      break;
    case 't': 
      trace_flag = atoi(&opt[2]);
      if(opt[strlen(opt)-1]=='m') 
        trace_flag*=1000000;
      if(trace_flag==0) trace_flag = 1;
      printf("*** gen trace data at %lld exec.\n",
             trace_flag);
      break;
    case 'v':
      verbose_interval =atoi(&opt[2]);
      if(verbose_interval == 0)   
	verbose_interval = 100000;
      if(opt[strlen(opt)-1]=='m') 
	verbose_interval*=1000000;
      printf("*** verbose interval: %d\n", 
	     verbose_interval);
      break;
    case 'f': 
      fd_adjust = atoi(&opt[2]);
      if(fd_adjust==0) fd_adjust=1;
      printf("*** fd_adjust=%d \n",fd_adjust);
      if(fd_adjust<0){
	printf("Error fd_adjust.\n");
	exit(1);
      }
      break;
    case 'e':
      end_tc =atoi(&opt[2]);
      if(opt[strlen(opt)-1]=='m')end_tc*=1000000;
      printf("** Finish at %lld code execution\n",
	     end_tc);
      break;
    case 'i':
      fd_stdin_flag = 1;
      fd_stdin = open(&opt[2], O_RDONLY);
      printf("*** SimAlpha: use [%s] as STDIN.\n", 
	     &opt[2]);
      if(fd_stdin<0){
	printf("Can not open file:%s\n",&opt[2]); 
	exit(1);
      }
      break;
    default:
      printf("*** Error:[%s] Wrong option!!\n\n",
	     opt);
      usage();
      exit(1);
    }
  }
  slow_mode = debug_flag || verbose_interval 
    || end_tc || trace_flag;
}

/************************************************/
evaluation_result::evaluation_result(){
  memset(this,0,sizeof(class evaluation_result));

  time_begin = time(0);
  gettimeofday(&tp, &tzp);
}
/************************************************/
